#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <curl/curl.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "grabber.h"
#include "util.h"

const char * const grabber_html_entities[][2] =
{
	{"thetasym", "ϑ"},
	{"Upsilon", "Υ"},
	{"upsilon", "υ"},
	{"Omicron", "Ο"},
	{"omicron", "ο"},
	{"Epsilon", "Ε"},
	{"epsilon", "ε"},
	{"alefsym", "ℵ"},
	{"Yacute", "Ý"},
	{"yacute", "ý"},
	{"weierp", "℘"},
	{"Ugrave", "Ù"},
	{"ugrave", "ù"},
	{"Uacute", "Ú"},
	{"uacute", "ú"},
	{"thinsp", " "},
	{"there4", "∴"},
	{"spades", "♠"},
	{"sigmaf", "ς"},
	{"Scaron", "Š"},
	{"scaron", "š"},
	{"rsaquo", "›"},
	{"rfloor", "⌋"},
	{"plusmn", "±"},
	{"permil", "‰"},
	{"otimes", "⊗"},
	{"Otilde", "Õ"},
	{"otilde", "õ"},
	{"Oslash", "Ø"},
	{"oslash", "ø"},
	{"Ograve", "Ò"},
	{"ograve", "ò"},
	{"Oacute", "Ó"},
	{"oacute", "ó"},
	{"Ntilde", "Ñ"},
	{"ntilde", "ñ"},
	{"middot", "·"},
	{"lsaquo", "‹"},
	{"lowast", "∗"},
	{"lfloor", "⌊"},
	{"Lambda", "Λ"},
	{"lambda", "λ"},
	{"iquest", "¿"},
	{"Igrave", "Ì"},
	{"igrave", "ì"},
	{"Iacute", "Í"},
	{"iacute", "í"},
	{"hellip", "…"},
	{"hearts", "♥"},
	{"frac34", "¾"},
	{"frac14", "¼"},
	{"frac12", "½"},
	{"forall", "∀"},
	{"Egrave", "È"},
	{"egrave", "è"},
	{"Eacute", "É"},
	{"eacute", "é"},
	{"divide", "÷"},
	{"Dagger", "‡"},
	{"dagger", "†"},
	{"curren", "¤"},
	{"Ccedil", "Ç"},
	{"ccedil", "ç"},
	{"brvbar", "¦"},
	{"Atilde", "Ã"},
	{"atilde", "ã"},
	{"Agrave", "À"},
	{"agrave", "à"},
	{"Aacute", "Á"},
	{"aacute", "á"},
	{"upsih", "ϒ"},
	{"Ucirc", "Û"},
	{"ucirc", "û"},
	{"trade", "™"},
	{"times", "×"},
	{"tilde", "˜"},
	{"THORN", "Þ"},
	{"thorn", "þ"},
	{"Theta", "Θ"},
	{"theta", "θ"},
	{"szlig", "ß"},
	{"Sigma", "Σ"},
	{"sigma", "σ"},
	{"sbquo", "‚"},
	{"rsquo", "’"},
	{"rdquo", "”"},
	{"rceil", "⌉"},
	{"raquo", "»"},
	{"radic", "√"},
	{"quot", "\""},
	{"Prime", "″"},
	{"prime", "′"},
	{"pound", "£"},
	{"oplus", "⊕"},
	{"Omega", "Ω"},
	{"omega", "ω"},
	{"oline", "‾"},
	{"OElig", "Œ"},
	{"oelig", "œ"},
	{"Ocirc", "Ô"},
	{"ocirc", "ô"},
	{"notin", "∉"},
	{"ndash", "–"},
	{"nabla", "∇"},
	{"minus", "−"},
	{"micro", "µ"},
	{"mdash", "—"},
	{"lsquo", "‘"},
	{"ldquo", "“"},
	{"lceil", "⌈"},
	{"laquo", "«"},
	{"Kappa", "Κ"},
	{"kappa", "κ"},
	{"infin", "∞"},
	{"image", "ℑ"},
	{"iexcl", "¡"},
	{"Icirc", "Î"},
	{"icirc", "î"},
	{"Gamma", "Γ"},
	{"gamma", "γ"},
	{"frasl", "⁄"},
	{"exist", "∃"},
	{"equiv", "≡"},
	{"empty", "∅"},
	{"Ecirc", "Ê"},
	{"ecirc", "ê"},
	{"diams", "♦"},
	{"Delta", "Δ"},
	{"delta", "δ"},
	{"crarr", "↵"},
	{"clubs", "♣"},
	{"cedil", "¸"},
	{"bdquo", "„"},
	{"asymp", "≈"},
	{"Aring", "Å"},
	{"aring", "å"},
	{"Alpha", "Α"},
	{"alpha", "α"},
	{"AElig", "Æ"},
	{"aelig", "æ"},
	{"acute", "´"},
	{"Acirc", "Â"},
	{"acirc", "â"},
	{"Zeta", "Ζ"},
	{"zeta", "ζ"},
	{"Yuml", "Ÿ"},
	{"yuml", "ÿ"},
	{"Uuml", "Ü"},
	{"uuml", "ü"},
	{"uArr", "⇑"},
	{"uarr", "↑"},
	{"supe", "⊇"},
	{"sup3", "³"},
	{"sup2", "²"},
	{"sup1", "¹"},
	{"sube", "⊆"},
	{"sect", "§"},
	{"sdot", "⋅"},
	{"real", "ℜ"},
	{"rArr", "⇒"},
	{"rarr", "→"},
	{"rang", "〉"},
	{"prop", "∝"},
	{"prod", "∏"},
	{"perp", "⊥"},
	{"part", "∂"},
	{"para", "¶"},
	{"Ouml", "Ö"},
	{"ouml", "ö"},
	{"ordm", "º"},
	{"ordf", "ª"},
	{"nsub", "⊄"},
	{"macr", "¯"},
	{"lArr", "⇐"},
	{"larr", "←"},
	{"lang", "〈"},
	{"Iuml", "Ï"},
	{"iuml", "ï"},
	{"isin", "∈"},
	{"Iota", "Ι"},
	{"iota", "ι"},
	{"hArr", "⇔"},
	{"harr", "↔"},
	{"fnof", "ƒ"},
	{"euro", "€"},
	{"Euml", "Ë"},
	{"euml", "ë"},
	{"ensp", " "},
	{"emsp", " "},
	{"dArr", "⇓"},
	{"darr", "↓"},
	{"copy", "©"},
	{"cong", "≅"},
	{"circ", "ˆ"},
	{"cent", "¢"},
	{"bull", "•"},
	{"Beta", "Β"},
	{"beta", "β"},
	{"Auml", "Ä"},
	{"auml", "ä"},
	{"apos", "'"},
	{"yen", "¥"},
	{"uml", "¨"},
	{"Tau", "Τ"},
	{"tau", "τ"},
	{"sup", "⊃"},
	{"sum", "∑"},
	{"sub", "⊂"},
	{"sim", "∼"},
	{"Rho", "Ρ"},
	{"rho", "ρ"},
	{"reg", "®"},
	{"Psi", "Ψ"},
	{"psi", "ψ"},
	{"piv", "ϖ"},
	{"Phi", "Φ"},
	{"phi", "φ"},
	{"not", "¬"},
	{"loz", "◊"},
	{"int", "∫"},
	{"ETH", "Ð"},
	{"eth", "ð"},
	{"Eta", "Η"},
	{"eta", "η"},
	{"deg", "°"},
	{"cup", "∪"},
	{"Chi", "Χ"},
	{"chi", "χ"},
	{"cap", "∩"},
	{"ang", "∠"},
	{"and", "∧"},
	{"amp", "&"},
	{"Xi", "Ξ"},
	{"xi", "ξ"},
	{"Pi", "Π"},
	{"pi", "π"},
	{"or", "∨"},
	{"Nu", "Ν"},
	{"nu", "ν"},
	{"ni", "∋"},
	{"ne", "≠"},
	{"Mu", "Μ"},
	{"mu", "μ"},
	{"lt", "<"},
	{"le", "≤"},
	{"gt", ">"},
	{"ge", "≥"},
	{0, 0}
};

size_t fetch_callback (void *ptr, size_t size, size_t nmemb, void *data)
{
	struct grabber_buffer *buffer = (struct grabber_buffer *) data;
	size_t ptrsize = size * nmemb;

	buffer->data = realloc (buffer->data, ptrsize + buffer->length + 1);
	if (buffer->data)
	{
		memcpy (&(buffer->data[buffer->length]), ptr, ptrsize);
		buffer->length += ptrsize;
		buffer->data[buffer->length] = 0;
	}

	return ptrsize;
}

struct grabber_handle *grabber_init (void)
{
	curl_global_init (CURL_GLOBAL_ALL);
	struct grabber_handle *h = malloc (sizeof (struct grabber_handle));
	h->curl = curl_easy_init ();
	if (!h->curl)
		return NULL;

	h->buffer.data = malloc (0);
	h->buffer.length = 0;

	curl_easy_setopt (h->curl, CURLOPT_WRITEFUNCTION, fetch_callback);
	curl_easy_setopt (h->curl, CURLOPT_WRITEDATA, (void *) &(h->buffer));
	curl_easy_setopt (h->curl, CURLOPT_USERAGENT, "libcurl/1.0");
	curl_easy_setopt (h->curl, CURLOPT_COOKIEJAR, NULL);
	curl_easy_setopt (h->curl, CURLOPT_COOKIELIST, "ALL");
	curl_easy_setopt (h->curl, CURLOPT_COOKIESESSION, 1l);

	return h;
}

void grabber_buffer_clean (struct grabber_buffer *buffer)
{
	buffer->length = 0;
	if (buffer->data)
		free (buffer->data);

	buffer->data = malloc (0);
}

unsigned int get_last_page (struct grabber_handle *handle)
{
	CURLcode ret;

	unsigned int last_page = 0;

	size_t urllen = 8 + strlen (XKCDSW_URL);
	char *url = malloc (urllen);
	snprintf (url, urllen, XKCDSW_URL, 0);
	url[urllen - 1] = 0;

	curl_easy_setopt (handle->curl, CURLOPT_URL, url);

	grabber_buffer_clean (&(handle->buffer));
	ret = curl_easy_perform (handle->curl);
	curl_easy_setopt (handle->curl, CURLOPT_URL, NULL);
	free (url);

	char *data = strdup (handle->buffer.data);
	grabber_buffer_clean (&(handle->buffer));

	if (ret != CURLE_OK)
		return 0;

	char *s = strstr (data, "<div class=\"pagination\">");
	if (!s)
	{
		free (data);
		return 0;
	}

	while (strlen (s) > 0 && *s != '\n')
		s++;

	char *s2 = strndup (s, 1024);
	free (data);
	data = NULL;
	s = NULL;

	s = strstr (s2, "Page <strong>");
	if (!s)
	{
		free (s2);
		return 0;
	}

	sscanf (s, "Page <strong>%*i</strong> of <strong>%i</strong>", &last_page);
	free (s2);

	return last_page;
}

size_t grabber_fetch_xkcdsw (struct grabber_handle *handle, size_t max_items, struct xkcdsw_item *item_buffer, unsigned int page_start)
{
	CURLcode ret;
	size_t num_found = 0;

	if (page_start == 0)
		page_start = 1;

	unsigned int post_start = 40 * (page_start - 1);
	size_t urllen = 8 + strlen (XKCDSW_URL);
	char *url = malloc (urllen);
	snprintf (url, urllen, XKCDSW_URL, post_start);
	url[urllen - 1] = 0;

	curl_easy_setopt (handle->curl, CURLOPT_URL, url);

	grabber_buffer_clean (&(handle->buffer));
	ret = curl_easy_perform (handle->curl);
	curl_easy_setopt (handle->curl, CURLOPT_URL, NULL);
	free (url);

	char *data = strdup (handle->buffer.data);
	grabber_buffer_clean (&(handle->buffer));

	if (ret != CURLE_OK)
		return 0;

	char *haystack = data;
	char *s;
	char imgformat[64], titletextformat[64];
	memset (imgformat, 0, 64);
	memset (titletextformat, 0, 64);
	sprintf (imgformat, "<img src=\"%%%is\"", URL_LENGTH - 1);
	sprintf (titletextformat, "<dd><em>%%%ic</em></dd>", TITLETEXT_LENGTH - 1);

	while ((s = strstr (haystack, "<dt class=\"attach-image\">")) != NULL && num_found < max_items)
	{
		s += 25;
		char imurl[URL_LENGTH];
		char titletext[TITLETEXT_LENGTH];
		memset (imurl, 0, URL_LENGTH);
		memset (titletext, 0, TITLETEXT_LENGTH);

		if (sscanf (s, imgformat, imurl) != 1)
		{
			haystack = NULL;
			s = NULL;
			free (data);
			return 0;
		}

		size_t len = strlen (imurl);
		if (imurl[len - 1] == '"')
			imurl[len - 1] = 0;

		s += 11 + strlen (imurl);

		while (strlen (s) != 0 && strncmp (s, "/></dt>", 7) != 0)
			s++;

		s += 7;
		char c = *s;
		while (c == '\n' || c == '\t' || c == '\r' || c == ' ')
		{
			s++;
			c = *s;
		}

		if (sscanf (s, titletextformat, titletext) == 1)
		{
			char *end = strstr (titletext, "</em></dd>");
			if (end == NULL)
			{
				fprintf (stderr, "No end tags in title text after %i characters - title text too long?\n", TITLETEXT_LENGTH - 1);
			}
			else
			{
				memset (end, 0, strlen (end) + 1);
				s += 18 + strlen (titletext);
			}
		}

		num_found++;

		char *imurl_repl = htmlent_replace (imurl);
		char *titletext_repl = htmlent_replace (titletext);

		char *sid_part = strstr (imurl_repl, "&sid=");
		if (sid_part && strlen (sid_part) >= 37)
			memset (sid_part, 0, 37);

		struct xkcdsw_item item;
		memset (item.url, 0, URL_LENGTH);
		memset (item.titletext, 0, TITLETEXT_LENGTH);
		if (strncmp (imurl_repl, "./", 2) == 0)
		{
			strcpy (item.url, XKCDSW_ATTACH_BASE);
			strcpy (&(item.url[strlen (item.url)]), &(imurl_repl[2]));
		}
		else
		{
			strcpy (item.url, imurl_repl);
		}
		if (strlen (titletext_repl) > 0)
			strcpy (item.titletext, titletext_repl);

		free (imurl_repl);
		free (titletext_repl);

		item_buffer[num_found - 1] = item;

		haystack = s;
	}

	haystack = NULL;
	s = NULL;
	free (data);

	return num_found;
}

void grabber_cleanup (struct grabber_handle *handle)
{
	curl_easy_cleanup (handle->curl);
	free (handle->buffer.data);
	handle->buffer.length = 0;

	free (handle);

	curl_global_cleanup ();
}

char *htmlent_replace (const char *s)
{
	size_t i, j, ptr = 0;
	char c;
	char *ret = calloc (2 * strlen (s) + 1, 1); /* This should be much more than enough. */

	for (i = 0; i < strlen (s); i++)
	{
		c = s[i];
		if (c != '&')
		{
			ret[ptr] = c;
			ptr++;
			continue;
		}
		else
		{
			for (j = 0; grabber_html_entities[j][0] != 0 && grabber_html_entities[j][1] != 0; j++)
			{
				const char * const *ent = grabber_html_entities[j];
				size_t ent_len = strlen (ent[0]);
				if (strncmp (&(s[i + 1]), ent[0], ent_len) == 0 && s[i + 1 + ent_len] == ';')
				{
					strcpy (&(ret[ptr]), ent[1]);
					ptr += strlen (ent[1]);
					i += ent_len + 1;
					break;
				}
			}
		}
	}

	size_t ret_len = strlen (ret);
	char *final_ret = malloc (ret_len + 1);
	strcpy (final_ret, ret);
	final_ret[ret_len] = 0;
	free (ret);

	return final_ret;
}

int download_comic (struct grabber_handle *handle, struct file_buffer *buf, const struct xkcdsw_item *item)
{
	CURLcode ret;

	curl_easy_setopt (handle->curl, CURLOPT_URL, item->url);

	grabber_buffer_clean (&(handle->buffer));
	ret = curl_easy_perform (handle->curl);
	curl_easy_setopt (handle->curl, CURLOPT_URL, NULL);

	if (ret != CURLE_OK)
		return 1;

	buf->length = handle->buffer.length;
	buf->data = malloc (buf->length); /* Note that we do not include the trailing \0. */
	memcpy (buf->data, handle->buffer.data, buf->length);

	grabber_buffer_clean (&(handle->buffer));

	memset (buf->ext, 0, FILEEXT_LENGTH);

	/* It's a kind of magic. */
	if (strncmp (buf->data, "\x47\x49\x46\x38\x39\x61", 6) == 0 ||
			strncmp (buf->data, "\x47\x49\x46\x38\x37\x61", 6) == 0)
		strcpy (buf->ext, ".gif");
	else if (strncmp (buf->data, "\xff\xd8", 2) == 0 &&
				strncmp (&(buf->data[buf->length - 2]), "\xff\xd9", 2) == 0)
		strcpy (buf->ext, ".jpg");
	else if (strncmp (buf->data, "\x89\x50\x4E\x47\x0D\x0A\x1A\x0A", 8) == 0)
		strcpy (buf->ext, ".png");

	return 0;
}
